/*
 * Cresques Mapping Suite. Graphic Library for constructing mapping applications.
 *
 * Copyright (C) 2004-5.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 * cresques@gmail.com
 */
package org.cresques.io;

import org.cresques.geo.ViewPortData;

import org.cresques.px.Extent;
import org.cresques.px.PxLayerList;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;


/**
 * Sirve datos solicitados por los drivers que salvan a raster. Implementa
 * IDataWriter que obliga al mtodo readData que es el que ser llamado
 * desde el driver cada vez que vacie el buffer y necesite ms datos.
 * @author Nacho Brodin (brodin_ign@gva.es)
 */
public class Rasterizer implements IDataWriter {
    private ViewPortData viewPort = null;
    private ViewPortData viewPortBlock = null;
    protected BufferedImage image;
    private PxLayerList layerList;
    protected int sizeBlock = 0;
    protected int lastBlock = 0;
    protected WritableRaster raster = null;
    protected int contBlocks = 1;
    protected double wcIntervalo = 0;
    protected int[] rasterData = null;
    protected Dimension dimension = null;
    protected int[] buffer;
    protected double wcAlto = 0;
    protected Graphics2D g;

    /**
     * Constructor para las clases que hereden de esta
     * y redefinan la funcionalidad que sirve los datos. Estas
     * tampoco usaran el viewPort ya que se supone que disponen
     * de su propia estructura para la seleccion de la vista
     */
    public Rasterizer(int sizeBlock) {
        this.sizeBlock = sizeBlock;
    }

    /**
     * Constructor
     * @param layerList
     * @param vp
     * @param sizeBlock
     */
    public Rasterizer(PxLayerList layerList, ViewPortData vp, int sizeBlock) {
        viewPort = (ViewPortData) vp.clone();

        //Calculo del viewPort del primer bloque
        wcAlto = viewPort.getExtent().maxY() - viewPort.getExtent().minY();
        wcIntervalo = (sizeBlock * wcAlto) / viewPort.getHeight();
        dimension = new Dimension((int) viewPort.getWidth(), sizeBlock);

        calcViewPort(viewPort);

        //Tamao de ltimo bloque
        int nBlocks = (int) (vp.getHeight() / sizeBlock);
        lastBlock = (int) (vp.getHeight() - (nBlocks * sizeBlock));

        this.layerList = layerList;
        this.sizeBlock = sizeBlock;

        //System.out.println("==>Iniciando servidor datos raster "+viewPort.getHeight()+" "+viewPort.getWidth()+" "+wcIntervalo +"..."+viewPortBlock.getExtent().maxX()+" - "+viewPortBlock.getExtent().maxY()+" - "+viewPortBlock.getExtent().minX()+" - "+viewPortBlock.getExtent().minY());
    }

    /**
     * Calculo del viewPort de cada bloque que vamos salvando a raster
     * @param vp        viewPort
     */
    private void calcViewPort(ViewPortData vp) {
        Extent ext = null;

        if (viewPortBlock == null) { //Primer bloque
            ext = new Extent(Math.round(vp.getExtent().minX()),
                             Math.round(vp.getExtent().maxY()),
                             Math.round(vp.getExtent().maxX()),
                             Math.round(vp.getExtent().maxY() - wcIntervalo));
        } else { //Bloques siguientes
            ext = new Extent(Math.round(vp.getExtent().minX()),
                             Math.round(vp.getExtent().minY()),
                             Math.round(vp.getExtent().maxX()),
                             Math.round(vp.getExtent().minY() - wcIntervalo));
        }

        viewPortBlock = new ViewPortData(vp.getProjection(), ext, dimension);
        viewPortBlock.zoom(ext);
    }

    /* (non-Javadoc)
     * @see org.cresques.io.IDataWriter#readData(int, int, int)
     */
    public int[] readData(int sX, int sY, int nBand) {
        if (nBand == 0) { //Con nBand==0 se devuelven las 3 bandas

            //image = new BufferedImage((int)this.viewPortBlock.getWidth(), (int)this.viewPortBlock.getHeight(), BufferedImage.TYPE_INT_RGB);
            image = new BufferedImage(sX, sY, BufferedImage.TYPE_INT_RGB);
            g = (Graphics2D) image.getGraphics();
            layerList.draw(g, viewPortBlock);

            rasterData = image.getRGB(0, 0, sX, sY, rasterData, 0, sX);

            //System.out.println("==>Cargando bloque ..."+(int)this.viewPortBlock.getWidth()+" * "+(int)this.viewPortBlock.getHeight()+" * "+viewPortBlock.getExtent().maxX()+" - "+viewPortBlock.getExtent().maxY()+" - "+viewPortBlock.getExtent().minX()+" - "+viewPortBlock.getExtent().minY());
            //System.out.println("==>Escribiendo linea ...0 - 0 - "+sX+" - "+sY+" - "+contBlocks);
            //Calculamos el viewPort del sgte bloque
            if (((contBlocks + 1) * sizeBlock) <= viewPort.getHeight()) {
                dimension = new Dimension(sX, sY);
            } else { //Calculo de la altura del ltimo bloque
                dimension = new Dimension(sX,
                                          (int) (viewPort.getHeight() -
                                          (contBlocks * sizeBlock)));
                wcIntervalo = (lastBlock * wcAlto) / viewPort.getHeight();
            }

            calcViewPort(viewPortBlock);

            contBlocks++;

            return rasterData;
        }

        return null;
    }

    /**
     * Asigna el ancho del bloque
     * @param sizeBlock Ancho del bloque en pixeles
     */
	public void setSizeBlock(int sizeBlock) {
		this.sizeBlock = sizeBlock;
	}
}
