package org.agil.core.spatialindex;

import java.io.*;

import java.util.*;

/**
 *  A page file that stores all pages into Persistent storage. It uses a
 *  RandomAccessFile to store node data. The format of the page file is as
 *  follows. First, a header is writen that stores important information about
 *  the RTree. The header format is as shown below: <br>
 *  &nbsp;&nbsp;int dimension<br>
 *  &nbsp;&nbsp;float fillFactor<br>
 *  &nbsp;&nbsp;int nodeCapacity<br>
 *  &nbsp;&nbsp;int pageSize<br>
 *  <p>
 *
 *  All the pages are stored after the header, with the following format: <br>
 *  &nbsp;&nbsp;int parent<br>
 *  &nbsp;&nbsp;int level<br>
 *  &nbsp;&nbsp;int usedSpace<br>
 *  &nbsp;&nbsp;// HyperCubes<br>
 *  &nbsp;&nbsp;for (i in usedSpace)<br>
 *  &nbsp;&nbsp;&nbsp;&nbsp;for (j in dimension) {<br>
 *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;float p(i)1 [j]<br>
 *  &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;float p(i)2 [j]<br>
 *  &nbsp;&nbsp;&nbsp;&nbsp;}<br>
 *  &nbsp;&nbsp;&nbsp;&nbsp;int branch<br>
 *  &nbsp;&nbsp;} <p>
 *
 *  Deleted pages are stored into a Stack. If a new entry is inserted it is
 *  placed in the last deleted page. That way the page file does not grow
 *  indefinetely. <p>
 *
 *  Created: Sun Oct 24 21:14:57 1999 <p>
 *
 *
 *
 *@author     Marios Hadjieleftheriou
 *@version    2
 */
public class PersistentPageFile extends PageFile {
	/**
	 *  Description of the Field
	 */
	protected String fileName;
	/**
	 *  Stores node data into Persistent storage.
	 */
	private RandomAccessFile file;
	private int active = 0;
	//para controlar la apertura/cierre del fichero
	private boolean isOpen = false;
	private Stack emptyPages = new Stack();
	/**
	 *  Header size calculated using the following formula: headerSize =
	 *  dimension + fillFactor + nodeCapacity + pageSize
	 */
	private int headerSize = 20;
	//16 - por compatibilidad con los
	//ficheros antiguos ponemos 20
			/**
	 *  Description of the Field
	 */
	public static final int EMPTY_PAGE = -2;


	/**
	 *  Constructor for the PersistentPageFile object
	 */
	public PersistentPageFile() {
		this(null);
	}


	/**
	 *  Constructor for the PersistentPageFile object
	 *
	 *@param  fileName  Description of Parameter
	 */
	public PersistentPageFile(String fileName) {
		try {
			synchronized (this) {
				active++;
				//            System.out.println("PersistentPageFile.active++:"+active);
				if (isOpen) {
					return;
				}
			}
			if (fileName == null) {
				File f = File.createTempFile("rtree", ".dat");
				this.fileName = f.getCanonicalPath();
				f.deleteOnExit();
				file = new RandomAccessFile(f, "rw");
			}
			else {
				//            System.out.println("Abriendo fichero " + fileName);
				isOpen = true;
				file = new RandomAccessFile(fileName, "rw");
				this.fileName = fileName;
				file.seek(0);
				byte[] header = new byte[headerSize];
				if (headerSize == file.read(header)) {
					DataInputStream ds = new DataInputStream(new ByteArrayInputStream(header));
					dimension = ds.readInt();
					fillFactor = ds.readFloat();
					nodeCapacity = ds.readInt();
					pageSize = ds.readInt();
					treeType = ds.readInt();
					//esto es por compatibilidad con el anterior fichero
					//System.out.println("Creando fichero " + fileName);
					//System.out.println("dimension: " + dimension + " fillFactor:" + fillFactor +
					//         " nodeCapacity:" + nodeCapacity + " pageSize:" + pageSize);
					// find all empty pages and add them into the emptyPages stack.
					long i = 1;
					try {
						while (true) {
							if (EMPTY_PAGE == file.readLong()) {
								emptyPages.push(new Long(i));
							}
							file.seek(headerSize + i * pageSize);
							i++;
						}
					}
					catch (IOException e) {
					}
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}


	/**
	 *  Creates an exact copy of this object.
	 *
	 *@return    Description of the Returned Value
	 */
	public Object clone() {
		return this;
	}


	/**
	 *  Description of the Method
	 *
	 *@param  tree        Description of Parameter
	 *@param  dimension   Description of Parameter
	 *@param  fillFactor  Description of Parameter
	 *@param  capacity    Description of Parameter
	 */
	protected void initialize(RTree tree, int dimension, float fillFactor, int capacity) {
		super.initialize(tree, dimension, fillFactor, capacity);
		emptyPages.clear();
		//System.out.println("Inicializando =>  pageSize:"+ pageSize);
		try {
			// FIXME: needs Java 2 to compile.
			file.setLength(0);
			file.seek(0);
			file.writeInt(dimension);
			file.writeFloat(fillFactor);
			file.writeInt(nodeCapacity);
			file.writeInt(pageSize);
			file.writeInt(this.tree.treeType);
		}
		catch (IOException e) {
			e.printStackTrace();
		}
	}


	/**
	 *  Overrides the finalize method to dispose of system resources or to
	 *  perform other cleanup when the PersistentPageFile object is garbage
	 *  collected.
	 *
	 *@exception  Throwable  Description of Exception
	 */
	protected void finalize() throws Throwable {
		try {
			file.close();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		super.finalize();
	}


	/**
	 *  Description of the Method
	 *
	 *@exception  IOException  Description of Exception
	 */
	protected void close() throws IOException {
		synchronized (this) {
			active--;
			//         System.out.println("PersistentPageFile.active--:"+active);
			if (active == 0) {
				//            System.out.println("Cerrando fichero " + this.fileName);
				file.close();
				isOpen = false;
			}
		}
	}


	/**
	 *  Description of the Method
	 *
	 *@param  page                Description of Parameter
	 *@return                     Description of the Returned Value
	 *@exception  PageFaultError  Description of Exception
	 */
	protected AbstractNode readNode(long page) throws PageFaultError {

		if (page < 0) {
			throw new IllegalArgumentException("Page number cannot be negative.");
		}

		byte[] b;
		int l;
		tree.stats.reads++;
		try {
			//System.out.println("Leyendo page:" + page + " Pos:" + (headerSize + page * pageSize) + " PageSize:" + pageSize);
			synchronized (this) {
				if (!isOpen) {
					return null;
				}
				active++;
				file.seek(headerSize + page * pageSize);
				b = new byte[pageSize];
				l = file.read(b);
			}
			/*
			 *  if (-1 == l) {
			 *  throw new PageFaultError("EOF found while trying to read page " + page + ".");
			 *  }
			 */
			DataInputStream ds = new DataInputStream(new ByteArrayInputStream(b));
			long parent = ds.readLong();
			/*
			 *  if (parent == EMPTY_PAGE) {
			 *  throw new PageFaultError("Page " + page + " is empty.");
			 *  }
			 */
			int level = ds.readInt();
			int usedSpace = ds.readInt();
			AbstractNode n;
			if (level != 0) {
				n = new Index(tree, parent, page, level);
			}
			else {
				n = new Leaf(tree, parent, page);
			}
			n.parent = parent;
			n.level = level;
			n.usedSpace = usedSpace;
			float[] p1 = new float[dimension];
			float[] p2 = new float[dimension];
			//System.out.println("page:" + page + " parent:" + parent + " level:" + level + " usedSpace:" + usedSpace);
			for (int i = 0; i < usedSpace; i++) {
				for (int j = 0; j < dimension; j++) {
					p1[j] = ds.readFloat();
					p2[j] = ds.readFloat();
				}
				n.data[i] = new HyperCube(new Point(p1), new Point(p2));
				n.branches[i] = ds.readLong();
				if (n.nodeMbb != null) {
					n.nodeMbb = n.nodeMbb.combinedMbb(n.data[i]);
				}
				else {
					n.nodeMbb = (HyperCube) n.data[i].clone();
				}
			}
			synchronized (this) {
				active--;
			}
			return n;
		}
		catch (IOException e) {
			System.out.println("Error al leer del fichero <" + this.fileName + ">");
			e.printStackTrace();
			synchronized (this) {
				active--;
			}
			return null;
		}
	}


	/**
	 *  Description of the Method
	 *
	 *@param  n                   Description of Parameter
	 *@return                     Description of the Returned Value
	 *@exception  PageFaultError  Description of Exception
	 */
	protected long writeNode(AbstractNode n) throws PageFaultError {
		/*
		 *  if (n == null) {
		 *  throw new IllegalArgumentException("Node cannot be null.");
		 *  }
		 */
		long page;
		try {
			synchronized (this) {
				if (!isOpen) {
					return -1;
				}
				tree.stats.writes++;
				active++;
				if (n.pageNumber < 0) {
					if (emptyPages.empty()) {
						page = (long) ((file.length() - headerSize) / pageSize);
					}
					else {
						page = ((Long) emptyPages.pop()).longValue();
					}
					n.pageNumber = page;
					tree.stats.pages++;
				}
				else {
					page = n.pageNumber;
				}
			}
			ByteArrayOutputStream bs = new ByteArrayOutputStream(pageSize);
			DataOutputStream ds = new DataOutputStream(bs);
			//System.out.println("Escribiendo page:" + page + " Pos:" + (headerSize + page * pageSize) + " PageSize:" + pageSize);
			//System.out.println("Parent:" + n.parent + " level:" + n.level + " usedSpace:" + n.usedSpace);
			ds.writeLong(n.parent);
			ds.writeInt(n.level);
			ds.writeInt(n.usedSpace);
			for (int i = 0; i < n.usedSpace; i++) {
				for (int j = 0; j < tree.getDimension(); j++) {
					ds.writeFloat(n.data[i].getLowPoint().getFloatCoordinate(j));
					ds.writeFloat(n.data[i].getHighPoint().getFloatCoordinate(j));
				}
				ds.writeLong(n.branches[i]);
			}
			for (int i = n.usedSpace; i < tree.getNodeCapacity(); i++) {
				for (int j = 0; j < tree.getDimension(); j++) {
					ds.writeFloat(Float.NaN);
					ds.writeFloat(Float.NaN);
				}
				ds.writeLong(-2);
			}
			ds.flush();
			bs.flush();
			synchronized (this) {
				file.seek(headerSize + page * pageSize);
				file.write(bs.toByteArray());
				active--;
			}
			return page;
		}
		catch (IOException e) {
			e.printStackTrace();
			active--;
			return -1;
		}
	}


	/**
	 *  Description of the Method
	 *
	 *@param  page                Description of Parameter
	 *@return                     Description of the Returned Value
	 *@exception  PageFaultError  Description of Exception
	 */
	protected AbstractNode deletePage(long page) throws PageFaultError {
		try {
			/*
			 *  if (page < 0 || page > (file.length() - headerSize) / pageSize) {
			 *  throw new PageFaultError("Page does not exist.");
			 *  } else {
			 */
			AbstractNode n = readNode(page);
			synchronized (this) {
				if (!isOpen) {
					return null;
				}
				active++;
				file.seek(headerSize + page * pageSize);
				file.writeLong(EMPTY_PAGE);
				emptyPages.push(new Long(page));
				tree.stats.pages--;
				active--;
			}
			return n;
			//}
		}
		catch (IOException e) {
			e.printStackTrace();
			synchronized (this) {
				active--;
			}
			return null;
		}
	}
}
// PersistentPageFile
